const deepEqual = require('fast-deep-equal');
const querystring = require('querystring');
const xhr = require('xhr');

const producerURL = 'https://www.varsnap.com/api/snap/produce/';
const consumerURL = 'https://www.varsnap.com/api/snap/consume/';
const version = 'v0.3.0';

const configKeys = {
  varsnap: 'varsnap',
  env: 'env',
  producerToken: 'producerToken',
  consumerToken: 'consumerToken',
};

class Util {
  static getConfig(config, configKey) {
    let val = config[configKey];
    if (!val) {
      return '';
    }
    val = val.toString().toLowerCase();
    return val;
  }

  static getSignature(targetFunc) {
    let signature = targetFunc.toString().split('{')[0];
    signature = signature.replace(/^\s+|\s+$/g, '');
    return 'javascript.' + version + '.' + signature;
  }

  static ajax(url, requestData) {
    const options = {
      method: 'POST',
      uri: url,
      body: querystring.stringify(requestData).toString('utf8'),
      headers: {'Content-Type': 'application/x-www-form-urlencoded'},
    };
    const ajaxPromise = new Promise((resolve, reject) => {
      xhr(options, (err, resp, body) => {
        if (err) {
          return reject(err);
        }
        const responseCode = parseInt(resp.statusCode / 100, 10);
        if (responseCode != 2) {
          return reject(new Error('Status: ' + resp.statusCode));
        }
        body = JSON.parse(body);
        return resolve(body);
      });
    });
    return ajaxPromise;
  }
}

class Producer {
  constructor(targetFunc, config) {
    this.targetFunc = targetFunc;
    this.config = config;
  }

  isEnabled() {
    if (Util.getConfig(this.config, configKeys.varsnap) !== 'true') {
      return false;
    }
    if (Util.getConfig(this.config, configKeys.env) !== 'production') {
      return false;
    }
    if (!Util.getConfig(this.config, configKeys.producerToken)) {
      return false;
    }
    return true;
  }

  static serialize(data) {
    return JSON.stringify(data);
  }

  produce(args, output) {
    if (!this.isEnabled()) {
      return;
    }
    console.log('Sending call to Varsnap'); // eslint-disable-line no-console
    const data = {
      producer_token: Util.getConfig(this.config, configKeys.producerToken),
      signature: Util.getSignature(this.targetFunc),
      inputs: Producer.serialize(args),
      prod_outputs: Producer.serialize(output),
    };
    Util.ajax(producerURL, data).catch((err) => {
      if(err.statusCode === 404) {
        console.error('Cannot produce to Varsnap; check your producer token'); // eslint-disable-line no-console
      } else {
        console.error(err.error); // eslint-disable-line no-console
      }
    });
  }
}

class Consumer {
  constructor(targetFunc, config) {
    this.targetFunc = targetFunc;
    this.config = config;
    this.lastSnapId = '';
  }

  isEnabled() {
    if (Util.getConfig(this.config, configKeys.varsnap) !== 'true') {
      return false;
    }
    if (Util.getConfig(this.config, configKeys.env) !== 'development') {
      return false;
    }
    if (!Util.getConfig(this.config, configKeys.consumerToken)) {
      return false;
    }
    return true;
  }

  static deserialize(data) {
    return JSON.parse(data);
  }

  consume() {
    if (!this.isEnabled()) {
      return;
    }
    console.log('Streaming in production Varsnap function calls...'); // eslint-disable-line no-console
    setInterval(this.consumeOnce.bind(this), 1000);
  }

  consumeOnce() {
    const data = {
      consumer_token: Util.getConfig(this.config, configKeys.consumerToken),
      signature: Util.getSignature(this.targetFunc),
    };
    Util.ajax(consumerURL, data).then((data) => {
      if(!data || data['id'] == this.lastSnapId) {
        return;
      }
      this.lastSnapId = data['id'];
      console.log('Receiving call from Varsnap uuid: ' + this.lastSnapId); // eslint-disable-line no-console
      const inputs = Consumer.deserialize(data['inputs']);
      const prodOutputs = Consumer.deserialize(data['prod_outputs']);
      const localOutputs = this.targetFunc(...inputs);
      this.report(inputs, prodOutputs, localOutputs);
    }).catch((error) => {
      console.error('Cannot contact Varsnap' + error); // eslint-disable-line no-console
    });
  }

  report(inputs, prodOutputs, localOutputs) {
    const functionName = Util.getSignature(this.targetFunc);
    console.log('Function:                      ' + functionName);   // eslint-disable-line no-console
    console.log('Function input args:           ' + inputs);         // eslint-disable-line no-console
    console.log('Production function outputs:   ' + prodOutputs);    // eslint-disable-line no-console
    console.log('Your function outputs:         ' + localOutputs);   // eslint-disable-line no-console
    const matches = deepEqual(prodOutputs, localOutputs);
    console.log('Matching outputs:              ' + matches);        // eslint-disable-line no-console
    console.log('');                                                 // eslint-disable-line no-console
  }
}

function varsnap(func) {
  const producer = new Producer(func, varsnap.config);
  const consumer = new Consumer(func, varsnap.config);

  consumer.consume();

  const wrapped = function(...args) {
    let result = '';
    try {
      result = func(...args);
    }
    catch(err) {
      producer.produce(args, err);
      throw(err);
    }
    producer.produce(args, result);
    return result;
  };

  wrapped.producer = producer;
  wrapped.consumer = consumer;
  return wrapped;
}

module.exports = {
  configKeys: configKeys,
  Util: Util,
  Producer: Producer,
  Consumer: Consumer,
  varsnap: varsnap,
  version: version,
};
